/*
 * Panel application "TETRIS"
 */
#include <application.h> 

/*
 * function prototypes
 */
int tetris_init(void);

const char tetris_name_P[] PROGMEM="Tetris";
const char tetris_icon_P[] PROGMEM={
  18,18,
  2,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x22,0x08,0x00,
  0x22,0x08,0x00,
  0x22,0x08,0x00,
  0x22,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x31,0x88,0x00,
  0x31,0xc8,0x00,
  0x39,0xe8,0x00,
  0x20,0xe8,0x00,
  0x3f,0xf8,0x00,

  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x09,0x00,
  0x20,0x09,0x80,
  0x20,0x09,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x20,0x08,0x00,
  0x31,0x98,0x00,
  0x31,0x38,0x00,
  0x3d,0x38,0x00,
  0x2e,0xe8,0x00,
  0x3f,0xf8,0x00
};

struct application TETRIS={tetris_name_P, tetris_icon_P, tetris_init, 0, 0};
typedef struct
{
	char x,y;
	u_char rotation;
	u_char type;
} PIECE;

// defines
#define ROT0   0 
#define ROT90  1
#define ROT180 2
#define ROT270 3

#define FIELD_W 15
#define FIELD_H 29

// some constants
const char title_P[]      PROGMEM = "TETRIS";
const char score_P[]      PROGMEM = "SCORE:";
const char level_P[]      PROGMEM = "LEVEL:";
const char game_P[]       PROGMEM = "GAME";
const char over_P[]       PROGMEM = "OVER";
const char PieceColor[8]  PROGMEM = {BLACK,RED,GREEN,YELLOW,RED,GREEN,YELLOW,RED};
const char PieceArray[8]  PROGMEM = {0x00, 0x33, 0x71, 0x17, 0x72, 0x63, 0x36, 0xf0};
const char RotArray[32]   PROGMEM = {0x00, 0x00, 0x00, 0x00,
                                     0x22, 0x22, 0x02, 0x20,
                                     0x11, 0x21, 0x02, 0x10,
                                     0x12, 0x11, 0x01, 0x20,
                                     0x11, 0x21, 0x02, 0x10,
                                     0x11, 0x11, 0x01, 0x10,
                                     0x11, 0x11, 0x01, 0x10,
                                     0x00, 0x10, 0x01, 0x00};

// globals
long tetrisscore;
long tetrishscore;
int tetrisx;
int tetrisy;
int tetrislevel;
int tetrisspeed;

/*
uchar Sound1[20]      = { 0x01, 0xF0, 0x07, 0xF0, 0x01, 0xF0, 0x05, 0xF0,
                0x01, 0xF0, 0x03, 0xF0, 0x01, 0xF0, 0x04, 0xF0,
                0x01, 0xF0, 0x06, 0xF0 };
uchar Sound2[20]      = { 0x01, 0xC0, 0x01, 0xD0, 0x01, 0xF0, 0x01, 0xA0,
                0x01, 0xF0, 0x01, 0xE0, 0x01, 0xB0, 0x01, 0xB0,
                0x01, 0xD0, 0x01, 0xA0 };
uchar Sound3[20]      = { 0x03, 0xA0, 0x04, 0xA0, 0x01, 0xD0, 0x01, 0xD0,
                0x05, 0xB0, 0x04, 0xC0, 0x01, 0xE0, 0x01, 0xE0,
                0x05, 0xC0, 0x02, 0xC0 };
uchar Sound4[20]      = { 0x02, 0xA0, 0x03, 0xA0, 0x03, 0xA0, 0x01, 0xA0,
                0x05, 0xA0, 0x05, 0xA0, 0x01, 0xA0, 0x03, 0xA0,
                0x07, 0xA0, 0x03, 0xA0 };
uchar Sound5[20]      = { 0x01, 0xFF, 0x01, 0xC2, 0x01, 0xA4, 0x01, 0x6A,
                0x01, 0xF0, 0x01, 0x89, 0x01, 0xD1, 0x01, 0xC7,
                0x01, 0xBB, 0x01, 0x6D };
*/

/*
 * draw a piece
 */
void showPiece(PIECE *piece, u_char color)
{
	u_char mask,bitmap, px, py, dx, dy;
	
	if (color!=BLACK) color=PRG_RDB(PieceColor+piece->type);
	bitmap=PRG_RDB(PieceArray+piece->type);
//	pw=PRG_RDB(PieceWidth+piece->type);
//	ph=PRG_RDB(PieceWidth+piece->type);
	dy=PRG_RDB(RotArray+(piece->type<<2)+piece->rotation);
	dx=dy>>4;
	dy&=0x0f;
	LEDNumout2(tetrislevel,2);
	LEDLongout(tetrisscore,8);
	switch (piece->rotation)
	{
		case ROT0:
	    mask=0x01;
	    for (py=0;py<2;py++)
	      for (px=0;px<4;px++)
	      {
		      if (bitmap&mask) LEDPixel(tetrisx+piece->x+px+dx-4,tetrisy+piece->y+py+dy-4,color);
		      mask<<=1;
	      }
	    break;  
		case ROT90:
	    mask=0x01;
	    for (py=0;py<2;py++)
	      for (px=0;px<4;px++)
	      {
		      if (bitmap&mask) LEDPixel(tetrisx+piece->x+1-py+dx-4,tetrisy+piece->y+px+dy-4,color);
		      mask<<=1;
	      }
	    break;  
		case ROT180:
	    mask=0x80;
	    for (py=0;py<2;py++)
	      for (px=0;px<4;px++)
	      {
		      if (bitmap&mask) LEDPixel(tetrisx+piece->x+px+dx-4,tetrisy+piece->y+py+dy-4,color);
		      mask>>=1;
	      }
	    break;  
		case ROT270:
	    mask=0x80;
	    for (py=0;py<2;py++)
	      for (px=0;px<4;px++)
	      {
		      if (bitmap&mask) LEDPixel(tetrisx+piece->x+1-py+dx-4,tetrisy+piece->y+px+dy-4,color);
		      mask>>=1;
	      }
	    break;  
  }  	
}

/*
 * check if piece is completely in the playfield and does not overlap with another piece
 */
u_char checkPiece(PIECE *piece)
{
	u_char mask,bitmap, px, py, dx, dy;
	
	bitmap=PRG_RDB(PieceArray+piece->type);
	dy=PRG_RDB(RotArray+(piece->type<<2)+piece->rotation);
	dx=dy>>4;
	dy&=0x0f;
	switch (piece->rotation)
	{
		case ROT0:
	    mask=0x01;
	    for (py=0;py<2;py++)
	      for (px=0;px<4;px++)
	      {
		      if ((bitmap&mask) && LEDGetPixel(tetrisx+piece->x+px+dx-4,tetrisy+piece->y+py+dy-4)) return 1;
		      mask<<=1;
	      }
	    break;  
		case ROT90:
	    mask=0x01;
	    for (py=0;py<2;py++)
	      for (px=0;px<4;px++)
	      {
		      if ((bitmap&mask) && LEDGetPixel(tetrisx+piece->x+1-py+dx-4,tetrisy+piece->y+px+dy-4)) return 1;
		      mask<<=1;
	      }
	    break;  
		case ROT180:
	    mask=0x80;
	    for (py=0;py<2;py++)
	      for (px=0;px<4;px++)
	      {
		      if ((bitmap&mask) && LEDGetPixel(tetrisx+piece->x+px+dx-4,tetrisy+piece->y+py+dy-4)) return 1;
		      mask>>=1;
	      }
	    break;  
		case ROT270:
	    mask=0x80;
	    for (py=0;py<2;py++)
	      for (px=0;px<4;px++)
	      {
		      if ((bitmap&mask) && LEDGetPixel(tetrisx+piece->x+1-py+dx-4,tetrisy+piece->y+px+dy-4)) return 1;
		      mask>>=1;
	      }
	    break;  
  }  	
  return 0;
}

void checkFullLines(void)
{
	u_char xs,ys,x,y,flag,n;
	n=0;
	for (ys=tetrisy;ys<tetrisy+FIELD_H;ys++)
	{
		flag=0;
		for (xs=tetrisx;xs<tetrisx+FIELD_W-1;xs++)
		  if (LEDGetPixel(xs,ys)) flag++;
		if (flag==FIELD_W-1)
		{
			n++;
			for (y=ys;y>tetrisy;y--)
			  for (x=tetrisx;x<tetrisx+FIELD_W-1;x++)
			    LEDPixel(x,y,LEDGetPixel(x,y-1));
		}	     
	}
	//printf("Full Lines: %i  Score: %li ",n,tetrisscore);
	switch (n)
	{
		case 1: tetrisscore+=10;break;
		case 2: tetrisscore+=25;break;
		case 3: tetrisscore+=50;break;
		case 4: tetrisscore+=100;break;
	}
	//printf("Score new: %li\n",tetrisscore);
}

u_char movePieceDown(PIECE *piece, PIECE *newpiece)
{
	showPiece(piece,BLACK);
	piece->y++;
  if (checkPiece(piece))
  {
  	piece->y--;
  	tetrisscore++;
  	showPiece(piece,RED);
    checkFullLines();
  	// create a new piece
  	piece->type=newpiece->type;
  	piece->rotation=newpiece->rotation;
  	piece->x=FIELD_W>>1;
  	piece->y=4;
 	  newpiece->x=FIELD_W+8;
 	  newpiece->y=tetrisy+5;
 	  newpiece->type=1+(char)rand()%7;
   	if (checkPiece(piece)) return 2;
  	showPiece(piece,RED);
  	LEDFRectangle(tetrisx+FIELD_W+4,tetrisy+3,4,4,BLACK);
  	showPiece(newpiece,RED);
   	LEDDraw();
  	return 0;
  }
  showPiece(piece,RED);
  LEDDraw();
  return 1;
}

u_char movePieceDownFast(PIECE *piece, PIECE *newpiece)
{
	u_char end;
	tetrisscore+=(FIELD_H-piece->y)>>1;
	while ((end=movePieceDown(piece,newpiece))==1) wait(1);
	return end;
}

void movePieceRight(PIECE *piece)
{
	showPiece(piece,BLACK);
	piece->x++;
	if (checkPiece(piece))
	{
		piece->x--;
		return;
	}	
	showPiece(piece,RED);
	LEDDraw();
}

void movePieceLeft(PIECE *piece)
{
	showPiece(piece,BLACK);
	piece->x--;
	if (checkPiece(piece))
	{
		piece->x++;
		return;
	}	
	showPiece(piece,RED);
	LEDDraw();
}

void rotatePiece(PIECE *piece)
{
 	showPiece(piece,BLACK);
	piece->rotation++;
	if (piece->rotation>ROT270) piece->rotation=ROT0;
	if (checkPiece(piece)) 
	{
	  if (piece->rotation)
	    piece->rotation--;
	  else
	    piece->rotation=ROT270;
	  return;  
	}
	showPiece(piece,RED);
	LEDDraw();
}

void TetrisGameOver(void)
{
	tetrislevel=0;
  LEDFStr_P(0, 10, MAXX, TE_CENTER, RED, game_P);
  LEDFStr_P(0, 20, MAXX, TE_CENTER, RED, over_P);
  LEDDraw();
  wait(3000);
}

void tetris_drawframe(void)
{
	char color=tetrislevel%3+1;
  LEDVLine(tetrisx-1,tetrisy,FIELD_H+1,color);
  LEDVLine(tetrisx+FIELD_W,tetrisy,FIELD_H+1,color);
  LEDHLine(tetrisx,tetrisy+FIELD_H,FIELD_W,color);		
}

int tetris_init(void)
{
	int pause;
	u_char keylock;
	PIECE piece, newpiece;
	
	// calculate screen position
	tetrisx=(MAXX-FIELD_W-4)>>1;
	tetrisy=(MAXY-FIELD_H)>>1;
	
  // initialize level
  pause=0;
  tetrislevel=1;
  tetrisspeed=500;
  tetrisscore=0;
  keylock=0;
	piece.x=FIELD_W>>1;
	piece.y=4;
	piece.rotation=ROT0;
	piece.type=1+(char)rand()%7;
  newpiece.x=FIELD_W+8;
  newpiece.y=tetrisy+5;;
	newpiece.rotation=ROT0;
	newpiece.type=1+(char)rand()%7;
	
	// draw the playfield
  LEDClear();
  tetris_drawframe();
  showPiece(&piece,RED);
  showPiece(&newpiece,RED);
  LEDDraw();
  
  // let the game begin
	do
	{
	  if (!keylock)
	  {
    	//if (UartKbhit() && (UartGetChar()==3)) return 0;
	  	if (keypad)
	  	{
	      if (keypad&BTN_LEFT)   movePieceLeft(&piece);
	      if (keypad&BTN_RIGHT)  movePieceRight(&piece);
	      if (keypad&BTN_DOWN) 	 if (movePieceDownFast(&piece,&newpiece)==2) TetrisGameOver();
		    if (keypad&BTN_UP)     rotatePiece(&piece);
	      if (keypad&BTN_ACTLEFT) return 0;
	      keylock=200;
	    }
	  }
	  else
	  {
	    if (!keypad) keylock=0;
	  	if (keylock) keylock--;
	  }  
    //while (keypad);
    pause++;
    if (pause>tetrisspeed)
    {
      if (tetrisscore>200*tetrislevel)
      {
      	tetrislevel++;
      	if (tetrislevel<11)
      	tetrisspeed=550-tetrislevel*50;
      	//printf("Speed: %i\n",tetrisspeed);
      	tetris_drawframe();
      	LEDDraw();
      }
    	if (movePieceDown(&piece,&newpiece)==2)
    	  TetrisGameOver();
    	pause=0;
    }
    if (wait(1)==3) return 0;
  } while (tetrislevel);
  return 0;
}

