/*
 * Panel application "Snake"
 */
#include <application.h>

/*
 * function prototypes
 */
int snake_init(void);

const char snake_name_P[] PROGMEM="Snake";
const char snake_icon_P[] PROGMEM={
  18,18,
  2,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x08,0x00,0x00,
  0x08,0x00,0x00,
  0x08,0x00,0x00,
  0x08,0x00,0x00,
  0x0f,0x80,0x00,
  0x00,0x80,0x00,
  0x00,0x80,0x00,
  0x00,0x80,0x00,
  0x00,0x80,0x00,
  0x00,0x80,0x00,
  0x00,0xf0,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,

  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x02,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x01,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x10,0x00,
  0x20,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00
};

struct application SNAKE={snake_name_P, snake_icon_P, snake_init, 0, 0};

#define SNAKE_MAXLEN       40
#define SNAKE_TAIL_COLOR   RED
#define SNAKE_HEAD_COLOR   GREEN
#define SNAKE_COOKIE_COLOR YELLOW
typedef struct
{
	char x;
	char y;
} SNAKEPIXEL;

/* This is the snake's pixels */
SNAKEPIXEL snake[SNAKE_MAXLEN];

/* The snakes head position in snake array */
int snake_head;

/* The snakes tail position in snake array */
int snake_tail;

/* The snake's current direction */
char snake_dir;

/* Snake Score counter */
long snake_score;

/* Time since last eating of a cookie */
int snake_time;

/*
 * create a new snake
 */
void snake_birth(void)
{
	char ox,oy;
	snake_score=0;
	snake_time=0;
	snake_head=0;
	snake_tail=0;
	snake_dir =0;
	snake[snake_head].x=MAXX>>1;
	snake[snake_head].y=MAXY>>1;
	LEDClear();
	LEDPixel(snake[snake_head].x,snake[snake_head].y,SNAKE_HEAD_COLOR);
	do  // draw a cookie
	{
	  ox=rand()%MAXX;
	  oy=rand()%MAXY;
	} while (LEDGetPixel(ox,oy)!=BLACK);
	LEDPixel(ox,oy,SNAKE_COOKIE_COLOR);
  LEDNumout2(0,0);
  LEDLongout(snake_score,10);
	LEDIntout(snake_time,3);
	LEDDraw();
}

/*
 * move the snake
 */
char snake_move(char dir)
{
	char ox,oy;
	// do not allow 180 degrees turns
	if (((dir==BTN_LEFT )&&(snake_dir==BTN_RIGHT))||
      ((dir==BTN_RIGHT)&&(snake_dir==BTN_LEFT ))||
      ((dir==BTN_UP   )&&(snake_dir==BTN_DOWN ))||
      ((dir==BTN_DOWN )&&(snake_dir==BTN_UP   )))
	  dir=0;
  // do only allow orthogonal directions
	if ((dir==BTN_LEFT)||(dir==BTN_RIGHT)||(dir==BTN_UP)||(dir==BTN_DOWN)) snake_dir=dir;
  if (!snake_dir) return 0;
  if (snake_time<10000) snake_time++;
	// old head becomes a tail now
	LEDPixel(snake[snake_head].x,snake[snake_head].y,SNAKE_TAIL_COLOR);
	ox=snake[snake_head].x;
	oy=snake[snake_head].y;
	snake_head=(snake_head+1)%SNAKE_MAXLEN;
	switch (snake_dir)
	{
		case BTN_LEFT:
	    if (ox>0)
	      ox--;
	    else
	      return 1;  // snake is dead
      break;
	  case BTN_RIGHT:
	    if (ox<MAXX-1)
	      ox++;
	    else
	      return 1;  // snake is dead
      break;
    case BTN_UP:
	    if (oy>0)
	      oy--;
	    else
	      return 1;  // snake is dead
	    break;
    case BTN_DOWN:
  	  if (oy<MAXY-1)
	      oy++;
	    else
	      return 1;  // snake is dead
	    break;
	}    
	snake[snake_head].x=ox;
	snake[snake_head].y=oy;
	// check if snake eats up itself
	ox=LEDGetPixel(snake[snake_head].x, snake[snake_head].y);
	if (ox==SNAKE_TAIL_COLOR) return 1; // snake is dead
	if (ox==BLACK) 
	{
		// snake simply moves, so cut of one tail pixel
	  LEDPixel(snake[snake_tail].x,snake[snake_tail].y,BLACK);
	  snake_tail=(snake_tail+1)%SNAKE_MAXLEN;
	}
	else
	{ // snake got a cookie
		// draw a new cookie
		do
		{
		  ox=rand()%MAXX;
		  oy=rand()%MAXY;
		} while (LEDGetPixel(ox,oy)!=BLACK);
		if (snake_time>100)
		  snake_score+=1;
		else
		  snake_score+=101-snake_time;
		snake_time=0;
		LEDPixel(ox,oy,SNAKE_COOKIE_COLOR);
	  LEDNumout2((snake_head>snake_tail?snake_head:snake_head+SNAKE_MAXLEN)-snake_tail+1,0);
	  LEDLongout(snake_score,10);
	}
	// add a new head pixel
	LEDPixel(snake[snake_head].x,snake[snake_head].y,SNAKE_HEAD_COLOR);
	LEDIntout(snake_time,3);
	LEDDraw();
	return 0;
}    	

/*
 * kill the snake, eat it up from the tail
 */
void snake_kill(void)
{
	while (snake_tail!=snake_head)
	{
	  LEDPixel(snake[snake_tail].x,snake[snake_tail].y,BLACK);
	  LEDDraw();
	  snake_tail=(snake_tail+1)%SNAKE_MAXLEN;
	  wait(20);
	}
}
	
int snake_init(void)
{
  snake_birth();
  do
  {
 	  if (snake_move(keypad))
 	  {
 	    snake_kill(); // snake is dead
 	    return 3;
 	  }
 	} while (wait(100)!=3);
  return 3;
}
