/*
 * Panel application "TIC TAC TOE"
 */
#include <application.h> 

/*
 * function prototypes
 */
int tictactoe_init(void);
const char tictactoe_name_P[] PROGMEM="TicTac";
const char tictactoe_icon_P[] PROGMEM=
{
  20,20,
  2,
  0x02,0x04,0x00,
  0x02,0x04,0x00,
  0x02,0x04,0x00,
  0x02,0x04,0x00,
  0x02,0x04,0x00,
  0x02,0x04,0x00,
  0xff,0xff,0xf0,
  0x02,0x04,0x00,
  0x02,0x95,0x20,
  0x02,0x64,0xc0,
  0x02,0x64,0xc0,
  0x02,0x95,0x20,
  0x02,0x04,0x00,
  0xff,0xff,0xf0,
  0x02,0x04,0x00,
  0x02,0x05,0x20,
  0x02,0x04,0xc0,
  0x02,0x04,0xc0,
  0x02,0x05,0x20,
  0x02,0x04,0x00,

  0x00,0x00,0x00,
  0x30,0x00,0x00,
  0x78,0x00,0x00,
  0x78,0x00,0x00,
  0x30,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x30,0x91,0x20,
  0x78,0x60,0xc0,
  0x78,0x60,0xc0,
  0x30,0x91,0x20,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x00,0x00,
  0x00,0x01,0x20,
  0x00,0x00,0xc0,
  0x00,0x00,0xc0,
  0x00,0x01,0x20,
  0x00,0x00,0x00
};

struct application TICTACTOE={tictactoe_name_P, tictactoe_icon_P, tictactoe_init, 0, 0};

// cross icon
const char tictactoe_cross_P[] PROGMEM=
{
  8,8,
  1,
  0xc3,0xe7,0x7e,0x3c,0x3c,0x7e,0xe7,0xc3
};

// circle icon
const char tictactoe_circle_P[] PROGMEM=
{
  8,8,
  1,
  0x3c,0x7e,0xe7,0xc3,0xc3,0xe7,0x7e,0x3c
};

// human icon
const char tictactoe_human_P[] PROGMEM=
{
  8,13,
  1,
  0x18,0x3c,0x3c,0x18,0xff,0xff,0x18,0x3c,0x3c,0x66,0x66,0xc3,0xc3
};

// machine icon
const char tictactoe_machine_P[] PROGMEM=
{
  13,13,
  1, 
  0x15,0x40,0x15,0x40,0x1f,0xe0,0xe0,0x38,0x28,0x20,0xe0,0x38,0x20,
  0x20,0xe0,0x38,0x20,0x20,0xe0,0x38,0x3f,0xe0,0x15,0x40,0x15,0x40
};




char tictactoe_field[3][3];
char tictactoe_mycolor=YELLOW;

void tictactoe_drawfield(void)
{
  int x,y;
  char x0=(MAXX-32)>>1; 
  char y0=(MAXY-32)>>1;
  LEDClear();
  LEDHLine(x0,10,y0+32,RED);
  LEDHLine(x0,21,y0+32,RED);
  LEDVLine(x0+10,y0+0,32,RED);
  LEDVLine(x0+21,y0+0,32,RED);
  for (y=0;y<3;y++)
    for (x=0;x<3;x++)
    {
      switch (tictactoe_field[x][y])
      {
        case BLACK : LEDFRectangle(x0+11*x,y0+11*y,10,10,BLACK); break;
        case YELLOW: LEDBitmap(x0+1+11*x,y0+1+11*y,YELLOW,tictactoe_cross_P); break;
        case GREEN : LEDBitmap(x0+1+11*x,y0+1+11*y,GREEN,tictactoe_circle_P); break;
      }  
    }
  LEDDraw();  
}

char tictactoe_checkwinner(void)
{
  char x0=(MAXX-32)>>1; 
  char y0=(MAXY-32)>>1;
  int i,j;
  // check for horizontal lines
  for (i=0;i<3;i++)
  {
    if ((tictactoe_field[0][i]!=BLACK) && (tictactoe_field[0][i]==tictactoe_field[1][i]) && (tictactoe_field[0][i]==tictactoe_field[2][i]))
    {
      // draw the winner line
      LEDHLine(x0,y0+4+11*i,32,tictactoe_field[0][i]);
      LEDHLine(x0,y0+5+11*i,32,tictactoe_field[0][i]);
      LEDDraw();
      wait(3000);
      return tictactoe_field[0][i];
    }
  } 

  // check for vertical lines
  for (i=0;i<3;i++)
  {
    if ((tictactoe_field[i][0]!=BLACK) && (tictactoe_field[i][0]==tictactoe_field[i][1]) && (tictactoe_field[i][0]==tictactoe_field[i][2]))
    {
      // draw the winner line
      for (j=0;j<5;j++)
      {
        LEDVLine(x0+4+11*i,y0,32,tictactoe_field[i][0]);
        LEDVLine(x0+5+11*i,y0,32,tictactoe_field[i][0]);
        LEDDraw();
        wait(100);
        tictactoe_drawfield();
        wait(100);
      }
      return tictactoe_field[i][0];
    }
  } 
  
  // check for diagonal lines
  if ((tictactoe_field[1][1]!=BLACK) && (tictactoe_field[0][0]==tictactoe_field[1][1]) && (tictactoe_field[1][1]==tictactoe_field[2][2]))
  {
    // draw the winner line
    for (j=0;j<5;j++)
    {
      LEDLine(x0,y0,x0+31,y0+31,tictactoe_field[1][1]);
      LEDLine(x0+1,y0,x0+31,y0+30,tictactoe_field[1][1]);
      LEDLine(x0,y0+1,x0+30,y0+31,tictactoe_field[1][1]);
      LEDDraw();
      wait(100);
      tictactoe_drawfield();
      wait(100);
    }
    return tictactoe_field[1][1];
  }
  if ((tictactoe_field[1][1]!=BLACK) && (tictactoe_field[2][0]==tictactoe_field[1][1]) && (tictactoe_field[1][1]==tictactoe_field[0][2]))
  {
    // draw the winner line
    for (j=0;j<5;j++)
    {
      LEDLine(x0,y0+31,x0+31,y0,tictactoe_field[1][1]);
      LEDLine(x0+1,y0+31,x0+31,y0+1,tictactoe_field[1][1]);
      LEDLine(x0,y0+30,x0+30,y0,tictactoe_field[1][1]);
      LEDDraw();
      wait(100);
      tictactoe_drawfield();
      wait(100);
    }
    return tictactoe_field[1][1];
  }
  
  // check for free fields
  for (i=0;i<3;i++)
    for (j=0;j<3;j++)
      if (tictactoe_field[i][j]==BLACK) return 0;
      
  // no winner  
  wait(3000);  
  return 99;
}

char tictactoe_human(char color)
{
  static int cx=0,cy=0;
  char x0=(MAXX-32)>>1; 
  char y0=(MAXY-32)>>1;

  tictactoe_drawfield();
  LEDRectangle(x0+11*cx,y0+11*cy,10,10,color);
  LEDDraw();
  do
  {
    if (UartKbhit() && (UartGetChar()==3)) return 0;
    if (keypad)
    {
      LEDRectangle(x0+11*cx,y0+11*cy,10,10,BLACK);
      if (keypad&BTN_LEFT)   {if (cx>0) cx--; else cx=2;}
      if (keypad&BTN_RIGHT)  {if (cx<2) cx++; else cx=0;}
      if (keypad&BTN_UP)     {if (cy>0) cy--; else cy=2;}
      if (keypad&BTN_DOWN)   {if (cy<2) cy++; else cy=0;}
      if (keypad&BTN_ACTION)  
      {
        if (tictactoe_field[cx][cy]==BLACK)
        {
          tictactoe_field[cx][cy]=color;
          tictactoe_drawfield();
          if (tictactoe_checkwinner()) return 0;
          while (keypad);
          return 1;
        }
      }
      if (keypad&BTN_ACTLEFT) return 0;
      LEDRectangle(x0+11*cx,y0+11*cy,10,10,color);
      LEDDraw();
      while (keypad);   
    }
  } while (1);
}   

/*
 * calculate a computer move
 */     
char tictactoe_computer(char color)
{ 
  int nx,ny,x,y,flag;
  char solutions[2][8];
  char mycolor=color;
  char encolor=(color==GREEN)?YELLOW:GREEN;
  
  tictactoe_drawfield();
  nx=ny=0;
  // check if field is empty
  flag=0;
  for (y=0;y<3;y++)
    for (x=0;x<3;x++) if (tictactoe_field[x][y]!=BLACK) flag=1;
  if (!flag)
  {
    switch (rand()%4)
    {
      case 0: nx=0; ny=0;break;
      case 1: nx=0; ny=2;break;
      case 2: nx=2; ny=0;break;
      case 3: nx=2; ny=2;break;
    }
    goto domove;
  }

  // calculate all solution entries
  // horizontal and vertical solutions
  memset(solutions,0,16);
  for (y=0;y<3;y++)
  {
    for (x=0;x<3;x++)
    {
      // horizontal
      solutions[0][y] +=  (tictactoe_field[x][y]==mycolor);
      solutions[1][y] +=  (tictactoe_field[x][y]==encolor);
      // vertical
      solutions[0][y+5] +=  (tictactoe_field[y][x]==mycolor);
      solutions[1][y+5] +=  (tictactoe_field[y][x]==encolor);
    }
    // diagonal
    solutions[0][3] += (tictactoe_field[y][y]==mycolor);
    solutions[1][3] += (tictactoe_field[y][y]==encolor);
    solutions[0][4] += (tictactoe_field[2-y][y]==mycolor);
    solutions[1][4] += (tictactoe_field[2-y][y]==encolor);
  }
  // output to console for testing
  printf("Solutions     myself enemy\n");
  for (x=0;x<3;x++) printf("Horizontal %i: %3i    %3i\n",x,solutions[0][x],solutions[1][x]);
  for (x=0;x<3;x++) printf("Vertical   %i: %3i    %3i\n",x,solutions[0][x+5],solutions[1][x+5]);
  printf("Diagonal   1: %3i    %3i\n",solutions[0][3],solutions[1][3]);
  printf("Diagonal   2: %3i    %3i\n",solutions[0][4],solutions[1][4]);

  // do we have two in line?
puts_P(PSTR("check if we have two in line"));
  x=0;
  do
  {    
    if ((solutions[0][x]==2) && (solutions[1][x]==0))
    {
puts_P(PSTR("check if we have two in line (hor)"));
      // horizontal
      if (x<3) 
      {
        ny=nx;
        if (tictactoe_field[0][ny]==BLACK) 
          nx=0;
        else
          if (tictactoe_field[1][ny]==BLACK) 
            nx=1;
          else
            nx=2;
        goto domove;    
      }
      // vertical
puts_P(PSTR("check if we have two in line (vert)"));
      if (x>4) 
      {
        nx=x-5;
        if (tictactoe_field[nx][0]==BLACK) 
          ny=0;
        else
          if (tictactoe_field[nx][1]==BLACK) 
            ny=1;
          else
            ny=2;
        goto domove;    
      }
      // diagonal
      if (x==3)
      {
        if (tictactoe_field[0][0]==BLACK) 
          nx=ny=0;
        else
          if (tictactoe_field[1][1]==BLACK) 
            nx=ny=1;
          else
            nx=ny=2;
        goto domove;    
      }
      if (tictactoe_field[2][0]==BLACK)
      {
        nx=2; ny=0;
      }  
      else
        if (tictactoe_field[1][1]==BLACK) 
          nx=ny=1;
        else
        {
          nx=0; ny=2;
        }  
      goto domove;
    }
    x++;
  } while (x<8);

  // does enemy have two in line?
puts_P(PSTR("check if enemy has two in line"));
  x=0;
  do
  {    
    if ((solutions[1][x]==2) && (solutions[0][x]==0))
    {
      // horizontal
puts_P(PSTR("check if enemy has two in line (hor)"));
      if (x<3) 
      {
        ny=x;
        if (tictactoe_field[0][ny]==BLACK) 
          nx=0;
        else
          if (tictactoe_field[1][ny]==BLACK) 
            nx=1;
          else
            nx=2;
        goto domove;    
      }
      // vertical
puts_P(PSTR("check if enemy has two in line (vert)"));
      if (x>4) 
      {
        nx=x-5;
        if (tictactoe_field[nx][0]==BLACK) 
          ny=0;
        else
          if (tictactoe_field[nx][1]==BLACK) 
            ny=1;
          else
            ny=2;
        goto domove;    
      }
      // diagonal
      if (x==3)
      {
        if (tictactoe_field[0][0]==BLACK) 
          nx=ny=0;
        else
          if (tictactoe_field[1][1]==BLACK) 
            nx=ny=1;
          else
            nx=ny=2;
        goto domove;    
      }
      if (tictactoe_field[2][0]==BLACK)
      {
        nx=2; ny=0;
      }  
      else
        if (tictactoe_field[1][1]==BLACK) 
          nx=ny=1;
        else
        {
          nx=0; ny=2;
        }  
      goto domove;  
    }  
    x++;
  } while (x<8);

  // try the center
puts_P(PSTR("try center"));
  if (tictactoe_field[1][1]==BLACK)
  {
    nx=ny=1;
    goto domove;
  }

  // check the corners for an blocking L move
puts_P(PSTR("check corners for blocking L move"));
  if ((tictactoe_field[0][0]==BLACK) && (tictactoe_field[1][0]==encolor) && (tictactoe_field[0][1]==encolor))
  {
    nx=ny=0;
    goto domove;
  }
  if ((tictactoe_field[2][0]==BLACK) && (tictactoe_field[1][0]==encolor) && (tictactoe_field[2][1]==encolor))
  {
    nx=2; ny=0;
    goto domove;
  }
  if ((tictactoe_field[0][2]==BLACK) && (tictactoe_field[0][1]==encolor) && (tictactoe_field[1][2]==encolor))
  {
    nx=0; ny=0;
    goto domove;
  }
  if ((tictactoe_field[2][2]==BLACK) && (tictactoe_field[2][1]==encolor) && (tictactoe_field[1][2]==encolor))
  {
    nx=ny=2;
    goto domove;
  }

  // if all corners are free get a random one
puts_P(PSTR("if all corners are free get a random one"));
  if ((tictactoe_field[0][0]==BLACK) && (tictactoe_field[2][0]==BLACK) &&
      (tictactoe_field[0][2]==BLACK) && (tictactoe_field[2][2]==BLACK))
  {
    switch (rand()%4)
    {
      case 0: nx=0; ny=0; break;
      case 1: nx=0; ny=2; break;
      case 2: nx=2; ny=0; break;
      case 3: nx=2; ny=2; break;
    }
    goto domove;
  }

  // check if an opposite corner of an enemyfield is free
puts_P(PSTR("check if an opposite corne of an enemyfield is free"));
  if ((tictactoe_field[0][0]==encolor) && (tictactoe_field[2][2]==BLACK))
  {
    nx=ny=2;
    goto domove;
  }
  if ((tictactoe_field[2][0]==encolor) && (tictactoe_field[0][2]==BLACK))
  {
    nx=0; ny=2;
    goto domove;
  }
  if ((tictactoe_field[0][2]==encolor) && (tictactoe_field[2][0]==BLACK))
  {
    nx=2; ny=0;
    goto domove;
  }
  if ((tictactoe_field[2][2]==encolor) && (tictactoe_field[0][0]==BLACK))
  {
    nx=ny=0;
    goto domove;
  }

  // do we have the center
puts_P(PSTR("do we have the center"));
  if (tictactoe_field[1][1]==mycolor)
  {
    if ((solutions[1][1]==1) && (solutions[0][1]==0))
    {
      nx=1;
      ny=(rand()&1)?0:2;
printf("My move1: %i %i\n",nx,ny);
      goto domove;
    }
    if ((solutions[1][6]==1) && (solutions[0][6]==0))
    {
      ny=1;
      nx=(rand()&1)?0:2;
printf("My move2: %i %i\n",nx,ny);
      goto domove;
    }
  }
  
  // if the center is lost get a corner
puts_P(PSTR("if center ist lost get a corner"));
  if (tictactoe_field[1][1]==encolor)
  {
    if (tictactoe_field[0][0]==BLACK)
    { 
      nx=ny=0;
      goto domove;
    }
    if (tictactoe_field[2][0]==BLACK)
    { 
      nx=2; ny=0;
      goto domove;
    }
    if (tictactoe_field[0][2]==BLACK)
    { 
      nx=0; ny=2;
      goto domove;
    }
    if (tictactoe_field[2][2]==BLACK)
    { 
      nx=ny=2;
      goto domove;
    }
  }
  
  // get a random free field
puts_P(PSTR("get a random free field"));
  do
  {
    flag=rand()%9;
    nx=flag/3; ny=flag%3;
  } while (tictactoe_field[nx][ny]!=BLACK);


domove:
  printf("My move: %i %i\n",nx,ny);
  // show the new part blinking a few times
  for (x=0;x<5;x++)
  {
    tictactoe_field[nx][ny]=BLACK;
    tictactoe_drawfield();
    wait(100);
    tictactoe_field[nx][ny]=color;
    tictactoe_drawfield();
    wait(100);
  }
  if (tictactoe_checkwinner()) return 0;
  return 1;
}
    
    
int tictactoe_init(void)
{
	char mode=0;
  puts_P(PSTR("Press Ctrl-C to quit"));
  
  do
  {
    LEDClear();
    switch (mode)
    {
    	case 0:
        LEDBitmap(2, 4,YELLOW,tictactoe_machine_P);
        LEDBitmap(MAXX-15,4,GREEN,tictactoe_machine_P);
        break;
      case 1:
        LEDBitmap(4, 4,YELLOW,tictactoe_human_P);
        LEDBitmap(MAXX-15,4,GREEN,tictactoe_machine_P);
        break;
      case 2:
        LEDBitmap(2, 4,YELLOW,tictactoe_machine_P);
        LEDBitmap(MAXX-12,4,GREEN,tictactoe_human_P);
        break;
      case 3:
        LEDBitmap(4, 4,YELLOW,tictactoe_human_P);
        LEDBitmap(MAXX-12,4,GREEN,tictactoe_human_P);
        break;
    }
    LEDFStr_P(0,MAXY-12, MAXX-1, TE_CENTER, RED,PSTR("Player"));
    LEDDraw();
    while(!keypad)
      if (UartKbhit() && (UartGetChar()==3)) return 0;
    if (keypad)
    {
      if (keypad&BTN_LEFT)   {if (mode>0) mode--; else mode=3;}
      if (keypad&BTN_RIGHT)  {if (mode<3) mode++; else mode=0;}
      if (keypad&BTN_ACTLEFT) return 0;
      while ((keypad) && (keypad!=BTN_ACTION)) ;
    }
  } while (!(keypad&BTN_ACTION));
  while (keypad);
  memset(tictactoe_field,BLACK,9);
  tictactoe_drawfield();
  do
  {
  	switch(mode)
  	{
  		case 0:
        if (!tictactoe_computer(YELLOW))
       	  memset(tictactoe_field,BLACK,9);
        if (wait(1000)==3) return 0;
        if (keypad&BTN_ACTLEFT) return 0;
        if (!tictactoe_computer(GREEN))
      	  memset(tictactoe_field,BLACK,9);
        if (wait(1000)==3) return 0;
        if (keypad&BTN_ACTLEFT) return 0;
        break;
      case 1:  
        if (!tictactoe_human(YELLOW))
       	  memset(tictactoe_field,BLACK,9);
        if (wait(1)==3) return 0;
        if (keypad&BTN_ACTLEFT) return 0;
        if (!tictactoe_computer(GREEN))
      	  memset(tictactoe_field,BLACK,9);
        if (wait(1)==3) return 0;
        if (keypad&BTN_ACTLEFT) return 0;
        break;
      case 2:  
        if (!tictactoe_computer(YELLOW))
       	  memset(tictactoe_field,BLACK,9);
        if (wait(1)==3) return 0;
        if (keypad&BTN_ACTLEFT) return 0;
        if (!tictactoe_human(GREEN))
      	  memset(tictactoe_field,BLACK,9);
        if (wait(1)==3) return 0;
        if (keypad&BTN_ACTLEFT) return 0;
        break;
      case 3:  
        if (!tictactoe_human(YELLOW))
       	  memset(tictactoe_field,BLACK,9);
        if (wait(1)==3) return 0;
        if (keypad&BTN_ACTLEFT) return 0;
        if (!tictactoe_human(GREEN))
      	  memset(tictactoe_field,BLACK,9);
        if (wait(1)==3) return 0;
        if (keypad&BTN_ACTLEFT) return 0;
        break;
    }
  } while (1);
  return 0;
}
