/*
 * Joystick Routines
 */
#include <avr/io.h>
#include <avr/signal.h>
#include "joystick.h"
#include "led.h"

#define UPPERLIMIT 220   // ADC value for UP or RIGHT key
#define LOWERLIMIT 30    // ADC value for DOWN or LEFT key

volatile JOYSTICK joystick;

unsigned char ADConvert(char chan)
{
  int tmp;
  int ADresult;

  ADMUX = (ADMUX & ~0x07)|chan;
  // NutDelay(40); // maybe delay a little here to settle the MUX
  for (tmp=0;tmp<10000;tmp++);
  sbi(ADCSRA,ADSC); // Start ADC 
  loop_until_bit_is_set(ADCSRA, ADIF); // Wait for ADIF, will happen soon
  // ADresult = ADCL; // Read LSB first
  // ADresult += ((int)ADCH) << 8; // then MSB
  ADresult = ADCH; // easier!!
  sbi(ADCSRA,ADIF); // Clear AD Interrupt Flag
  return ADresult;
}


SIGNAL(SIG_ADC)
{
	static char buttons;
	static char wait=3;
	static char chan=2;
	char val;
  val=ADCH;
  if (wait)
    wait--;
  else
  {
    joystick.buttons=~(PINF|joystick.mode|1);
    switch (chan)
    {
    	case 2: 
    	  if (joystick.mode & 0x04) 
    	  {
    	    buttons&=~(BTN_RIGHT|BTN_LEFT);
    	    joystick.x=val;
    	    if (val>UPPERLIMIT) 
    	      buttons|=BTN_RIGHT;
    	    else if (val<LOWERLIMIT)
    	      buttons|=BTN_LEFT;
    	    joystick.buttons|=buttons;  
        }
    	  else
    	  {
    	    if ((joystick.buttons&BTN_RIGHT) && (joystick.x<255)) joystick.x++;
    	    if ((joystick.buttons&BTN_LEFT) && (joystick.x>0)) joystick.x--;
    	  }
    	  chan=3;
    	  break;
    	case 3: 
    	  if (joystick.mode & 0x08)
    	  {
     	  	buttons&=~(BTN_UP|BTN_DOWN);
    	    joystick.y=val;
    	    if (val>UPPERLIMIT) 
    	      buttons|=BTN_UP;
    	    else if (val<LOWERLIMIT)
    	      buttons|=BTN_DOWN;
    	    joystick.buttons|=buttons;  
    	  }   
    	  else
    	  {
    	    if ((joystick.buttons&BTN_UP) && (joystick.y<255)) joystick.y++;
    	    if ((joystick.buttons&BTN_DOWN) && (joystick.y>0)) joystick.y--;
    	  }
    	  chan=2;
    	  break;
    }
    ADMUX = (ADMUX & ~0x07)|chan;
    wait=2;
  }
  //ADCSRA = _BV(ADEN) + _BV(ADIE) + _BV(ADSC); // Start ADC 
  sbi(ADCSRA,ADSC);
}	

/*
 * read Joystick state
 *
 * x,y    : Paddle x and y position
 * buttons: BTN_LEFT, BTN_RIGHT, BTN_UP, BTN_DOWN, BTN_ACTION, BTN_ACTLEFT, BTN_ACTRIGHT
 */ 
void joystickRead(unsigned char *x, unsigned char *y, unsigned char *buttons)
{
	*x=ADConvert(2);
	*y=ADConvert(3);
	*buttons=~PINF&0xF2;
}


void joystickInit(void)
{
	unsigned char x; 
  DDRF   = 0x00;  // DataDirectionRegister for PortF to input
  PORTF  = 0xfe; // first switch all pullups on
  ADCSRA = _BV(ADEN);
  ADMUX  = _BV(REFS0) + _BV(ADLAR); // AVCC as reference voltage, left justified // orig. 0xC0

	x=ADConvert(2); // one dummy read for initializing
  joystick.mode=0;
  // read the x ADC
	x=ADConvert(2);
  if (x<240) // look if there is a poti connected
  {
  	joystick.mode=0x04;
  	DDRF&=0xfc;  // switch pullup off
  }
  
  // read the y ADC
	x=ADConvert(3);
  if (x<240) // look if there is a poti connected
  {
  	joystick.mode|=0x08;
  	DDRF&=0xf7;  // switch pullup off
  }
  ADCSRA = _BV(ADEN) + _BV(ADIE) + _BV(ADPS1) + _BV(ADPS2);// + _BV(ADFR);
  sbi(ADCSRA,ADSC); // Start ADC 
}
